@@ -38,7 +38,8 @@
 char *progname;
 static struct hd_driveid id;
 static int verbose = 0, get_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
-static int flagcount = 0, do_flush = 0, do_timings = 0, is_scsi_hd = 0;
+static int flagcount = 0, do_flush = 0, is_scsi_hd = 0;;
+static int do_ctimings, do_timings = 0;
 
 static unsigned long set_readahead= 0, get_readahead= 0, readahead= 0;     
 static unsigned long set_readonly = 0, get_readonly = 0, readonly = 0;     
@@ -199,7 +200,7 @@ int seek_to_zero (int fd)
 
 int read_big_block (int fd, char *buf)
 {
-	int rc;
+	int i, rc;
 	if ((rc = read(fd, buf, TIMING_BUF_BYTES)) != TIMING_BUF_BYTES) {
 		if (rc)
 			perror("read() failed");
@@ -207,14 +208,18 @@ int read_big_block (int fd, char *buf)
 			fputs ("read() hit EOF - device too small\n", stderr);
 		return 1;
 	}
+	/* access all sectors of buf to ensure the read fully completed */
+	for (i = 0; i < TIMING_BUF_BYTES; i += 512)
+		buf[i] &= 1;
 	return 0;
 }
 
-void time_device (int fd)
+static double correction = 0.0;
+
+void time_cache (int fd)
 {
 	int i;
 	char *buf;
-	double elapsed, correction;
 	struct itimerval e1, e2;
 	int shmid;
 
@@ -251,9 +256,6 @@ void time_device (int fd)
 	setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
 	if (seek_to_zero (fd)) return;
 	if (read_big_block (fd, buf)) return;
-	/* access all sectors of buf to ensure the read fully completed */
-	for (i = 0; i < TIMING_BUF_BYTES; i += 512)
-		buf[i] &= 1;
 	printf(" Timing buffer-cache reads:   ");
 	fflush(stdout);
 
@@ -287,10 +289,49 @@ void time_device (int fd)
 
 	flush_buffer_cache(fd);
 	sleep(1);
+quit:
+	if (-1 == shmdt(buf))
+		perror ("could not detach sharedmem buf");
+}
+
+void time_device (int fd)
+{
+	int i;
+	char *buf;
+	double elapsed;
+	struct itimerval e1, e2;
+	int shmid;
+
+	if ((shmid = shmget(IPC_PRIVATE, TIMING_BUF_BYTES, 0600)) == -1) {
+		perror ("could not allocate sharedmem buf");
+		return;
+	}
+	if (shmctl(shmid, SHM_LOCK, NULL) == -1) {
+		perror ("could not lock sharedmem buf");
+		(void) shmctl(shmid, IPC_RMID, NULL);
+		return;
+	}
+	if ((buf = shmat(shmid, (char *) 0, 0)) == (char *) -1) {
+		perror ("could not attach sharedmem buf");
+		(void) shmctl(shmid, IPC_RMID, NULL);
+		return;
+	}
+	if (shmctl(shmid, IPC_RMID, NULL) == -1)
+		perror ("shmctl(,IPC_RMID,) failed");
+
+	/* Clear out the device request queues & give them time to complete */
+	sync();
+	sleep(3);
 
 	printf(" Timing buffered disk reads:  ");
 	fflush(stdout);
 
+	/*
+	 * getitimer() is used rather than gettimeofday() because
+	 * it is much more consistent (on my machine, at least).
+	 */
+	setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
+
 	/* Now do the timings for real */
 	getitimer(ITIMER_REAL, &e1);
 	for (i = TIMING_BUF_COUNT; i > 0; --i) {
@@ -306,7 +347,7 @@ void time_device (int fd)
 	if (elapsed <= (correction * 2))
 		printf("Hmm.. suspicious results: probably not enough free memory for a proper test.\n");
 #if 0  /* the "estimate" is just plain wrong for many systems.. */
-	else {
+	else if (correction != 0.0) {
 		printf(" Estimating raw driver speed: ");
 		elapsed -= correction;
 		printf("%2d MB in %5.2f seconds =%5.2f MB/sec\n",
@@ -750,6 +791,8 @@ void process_dev (char *devname)
 			perror(" HDIO_DRIVE_CMD failed");
 	}
 #endif
+	if (do_ctimings)
+		time_cache (fd);
 	if (do_timings)
 		time_device (fd);
 	if (do_flush)
@@ -759,7 +802,7 @@ void process_dev (char *devname)
 
 void usage_error (void)
 {
-	fprintf(stderr,"\n%s - get/set hard disk parameters - version 2.9\n\n", progname);
+	fprintf(stderr,"\n%s - get/set hard disk parameters - version 3.1\n\n", progname);
 	fprintf(stderr,"Usage:  %s  [options] [device] ..\n\n", progname);
 	fprintf(stderr,"Options:\n"
 	" -a    get/set fs readahead\n"
@@ -801,6 +844,7 @@ void usage_error (void)
 	" -S  * set standby (spindown) timeout\n"
 #endif
 	" -t    perform device read timings\n"
+	" -T    perform cache read timings\n"
 	" -u  * get/set unmaskirq flag (0/1) (DANGEROUS)\n"
 	" -v    default; same as -acdgkmnru (-gr for SCSI)\n"
 #ifdef HDIO_DRIVE_CMD
@@ -901,7 +945,7 @@ int main(int argc, char **argv)
 						noisy = 1;
 						if (!*p && argc && isdigit(**argv))
 							p = *argv++, --argc;
-						if (*p >= '0' && *p <= '5')
+						if (*p >= '0' && *p <= '9')
 							piomode = *p++ - '0';
 						else
 							piomode = 255;  /* auto-tune */
@@ -1017,6 +1061,9 @@ int main(int argc, char **argv)
 					case 't':
 						do_timings = 1;
 						break;
+					case 'T':
+						do_ctimings = 1;
+						break;
 					case 'h':
 					default:
 						usage_error();
